import logging
import math
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.constants import (
    PYARROW_INFLATION_MULTIPLIER,
    BYTES_PER_MEBIBYTE,
    MEMORY_TO_HASH_BUCKET_COUNT_RATIO,
)

from deltacat.storage import (
    PartitionLocator,
    Delta,
    ManifestEntry,
    interface as unimplemented_deltacat_storage,
)
from deltacat import logs
from deltacat.compute.compactor import DeltaAnnotated
from typing import Dict, List, Optional, Tuple, Union, Any
from deltacat.compute.compactor import HighWatermark
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def discover_deltas(
    source_partition_locator: PartitionLocator,
    high_watermark: Union[HighWatermark, int],
    last_stream_position_to_compact: int,
    compacted_partition_locator: Optional[PartitionLocator],
    rebase_source_partition_locator: Optional[PartitionLocator],
    rebase_source_partition_high_watermark: Optional[int],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
    list_deltas_kwargs: Optional[Dict[str, Any]] = {},
) -> Tuple[List[Delta], int]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    # Source One: new deltas from uncompacted table for incremental compaction or deltas from compacted table for rebase
    start_position_exclusive = (
        high_watermark.get(source_partition_locator)
        if isinstance(high_watermark, dict)
        else high_watermark
    )
    input_deltas = _discover_deltas(
        source_partition_locator,
        start_position_exclusive,
        last_stream_position_to_compact
        if not rebase_source_partition_locator
        else deltacat_storage.get_partition(
            source_partition_locator.stream_locator,
            source_partition_locator.partition_values,
            **deltacat_storage_kwargs,
        ).stream_position,
        deltacat_storage,
        deltacat_storage_kwargs,
        list_deltas_kwargs,
    )

    # Source Two: delta from compacted table for incremental compaction or new deltas from uncompacted table for rebase
    previous_last_stream_position_compacted = -1
    if not rebase_source_partition_locator:  # compacted table
        compacted_partition = deltacat_storage.get_partition(
            compacted_partition_locator.stream_locator,
            compacted_partition_locator.partition_values,
            **deltacat_storage_kwargs,
        )
        previous_last_stream_position_compacted = (
            compacted_partition.stream_position if compacted_partition else -1
        )
        input_deltas_compacted = []
        if previous_last_stream_position_compacted > 0:
            input_deltas_compacted = _discover_deltas(
                compacted_partition_locator,
                None,
                previous_last_stream_position_compacted,
                deltacat_storage,
                deltacat_storage_kwargs,
                list_deltas_kwargs,
            )
        logger.info(
            f"Length of input deltas from uncompacted table {len(input_deltas)} up to {last_stream_position_to_compact},"
            f"Length of input deltas from compacted table {len(input_deltas_compacted)} up to {high_watermark}"
        )
        input_deltas += input_deltas_compacted
    else:  # new deltas from uncompacted table between previous_last_stream_position_compacted and current last_position_to_compact
        input_deltas_new = _discover_deltas(
            rebase_source_partition_locator,
            rebase_source_partition_high_watermark,
            last_stream_position_to_compact,
            deltacat_storage,
            deltacat_storage_kwargs,
            list_deltas_kwargs,
        )
        logger.info(
            f"Length of input deltas from uncompacted table {len(input_deltas_new)} up to {last_stream_position_to_compact},"
            f"Length of input deltas from compacted table {len(input_deltas)} up to {rebase_source_partition_high_watermark}"
        )
        input_deltas += input_deltas_new

    return input_deltas, previous_last_stream_position_compacted


def limit_input_deltas(
    input_deltas: List[Delta],
    cluster_resources: Dict[str, float],
    hash_bucket_count: int,
    user_hash_bucket_chunk_size: int,
    input_deltas_stats: Dict[int, DeltaStats],
    compaction_audit: CompactionSessionAuditInfo,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[List[DeltaAnnotated], int, HighWatermark, bool]:
    # TODO (pdames): when row counts are available in metadata, use them
    #  instead of bytes - memory consumption depends more on number of
    #  input delta records than bytes.

    # we assume here that we're running on a fixed-size cluster
    # this assumption could be removed, but we'd still need to know the max
    # resources we COULD get for this cluster, and the amount of memory
    # available per CPU should remain fixed across the cluster.
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    worker_cpus = int(cluster_resources["CPU"])
    worker_obj_store_mem = float(cluster_resources["object_store_memory"])
    logger.info(f"Total worker object store memory: {worker_obj_store_mem}")
    worker_obj_store_mem_per_task = worker_obj_store_mem / worker_cpus
    logger.info(f"Worker object store memory/task: " f"{worker_obj_store_mem_per_task}")
    worker_task_mem = cluster_resources["memory"]
    logger.info(f"Total worker memory: {worker_task_mem}")
    # TODO (pdames): ensure fixed memory per CPU in heterogenous clusters
    worker_mem_per_task = worker_task_mem / worker_cpus
    logger.info(f"Cluster worker memory/task: {worker_mem_per_task}")

    delta_bytes = 0
    delta_bytes_pyarrow = 0
    delta_manifest_entries = 0
    require_multiple_rounds = False
    # tracks the latest stream position for each partition locator
    high_watermark = HighWatermark()
    limited_input_da_list = []

    if input_deltas_stats is None:
        input_deltas_stats = {}

    input_deltas_stats = {
        int(stream_pos): DeltaStats(delta_stats)
        for stream_pos, delta_stats in input_deltas_stats.items()
    }
    for delta in input_deltas:
        manifest = deltacat_storage.get_delta_manifest(delta, **deltacat_storage_kwargs)
        delta.manifest = manifest
        position = delta.stream_position
        delta_stats = input_deltas_stats.get(delta.stream_position, DeltaStats())
        if delta_stats:
            # TODO (pdames): derive from row count instead of table bytes
            delta_bytes_pyarrow += delta_stats.stats.pyarrow_table_bytes
        else:
            # TODO (pdames): ensure pyarrow object fits in per-task obj store mem
            logger.warning(
                f"Stats are missing for delta stream position {delta.stream_position}, "
                f"materialized delta may not fit in per-task object store memory."
            )
        manifest_entries = delta.manifest.entries
        delta_manifest_entries += len(manifest_entries)
        for entry in manifest_entries:
            delta_bytes += entry.meta.content_length
            if not delta_stats:
                delta_bytes_pyarrow = delta_bytes * PYARROW_INFLATION_MULTIPLIER
        high_watermark.set(
            delta.locator.partition_locator,
            max(position, high_watermark.get(delta.locator.partition_locator)),
        )
        if delta_bytes_pyarrow > worker_obj_store_mem:
            logger.info(
                f"Input deltas limited to "
                f"{len(limited_input_da_list)} by object store mem "
                f"({delta_bytes_pyarrow} > {worker_obj_store_mem})"
            )
            require_multiple_rounds = True
            break
        delta_annotated = DeltaAnnotated.of(delta)
        limited_input_da_list.append(delta_annotated)

    logger.info(f"Input deltas to compact this round: " f"{len(limited_input_da_list)}")
    logger.info(f"Input delta bytes to compact: {delta_bytes}")
    logger.info(f"Input delta files to compact: {delta_manifest_entries}")
    logger.info(f"Latest input delta stream position: {high_watermark}")

    if not limited_input_da_list:
        raise RuntimeError("No input deltas to compact!")

    # TODO (pdames): determine min hash buckets from size of all deltas
    #  (not just deltas for this round)
    min_hash_bucket_count = int(
        max(
            math.ceil(delta_bytes_pyarrow / worker_obj_store_mem_per_task),
            min(worker_cpus, 256),
        )
    )
    logger.info(f"Minimum recommended hash buckets: {min_hash_bucket_count}")

    if hash_bucket_count is None:
        # TODO (pdames): calc default hash buckets from table growth rate... as
        #  this stands, we don't know whether we're provisioning insufficient
        #  hash buckets for the next 5 minutes of deltas or more than enough
        #  for the next 10 years
        hash_bucket_count = min_hash_bucket_count
        logger.info(f"Using default hash bucket count: {hash_bucket_count}")

    if hash_bucket_count < min_hash_bucket_count:
        logger.warning(
            f"Provided hash bucket count ({hash_bucket_count}) "
            f"is less than the min recommended ({min_hash_bucket_count}). "
            f"This compaction job run may run out of memory, or run slowly. To "
            f"resolve this problem either specify a larger number of hash "
            f"buckets when running compaction, omit a custom hash bucket "
            f"count when running compaction, or provision workers with more "
            f"task memory per CPU."
        )

    hash_bucket_chunk_size = user_hash_bucket_chunk_size
    max_hash_bucket_chunk_size = math.ceil(
        worker_obj_store_mem_per_task / PYARROW_INFLATION_MULTIPLIER
    )
    logger.info(f"Max hash bucket chunk size: {max_hash_bucket_chunk_size}")
    if hash_bucket_chunk_size > max_hash_bucket_chunk_size:
        # TODO (pdames): note type of memory to increase (task or object store)
        logger.warning(
            f"Provided hash bucket chunk size "
            f"({user_hash_bucket_chunk_size}) is greater than the max "
            f"recommended ({max_hash_bucket_chunk_size}). This compaction "
            f"job may run out of memory. To resolve this problem either "
            f"specify a smaller hash bucket chunk size when running "
            f"compaction, omit a custom hash bucket chunk size when running "
            f"compaction, or provision workers with more task and object "
            f"store memory per CPU."
        )
    elif not hash_bucket_chunk_size:
        hash_bucket_chunk_size_load_balanced = max(
            math.ceil(max(delta_bytes, delta_bytes_pyarrow) / worker_cpus),
            BYTES_PER_MEBIBYTE,
        )
        hash_bucket_chunk_size = min(
            max_hash_bucket_chunk_size,
            hash_bucket_chunk_size_load_balanced,
        )
        logger.info(f"Default hash bucket chunk size: {hash_bucket_chunk_size}")

    rebatched_da_list = DeltaAnnotated.rebatch(
        limited_input_da_list,
        hash_bucket_chunk_size,
        # TODO (pdames): Test and add value for min_file_counts
    )

    compaction_audit.set_input_size_bytes(delta_bytes)
    compaction_audit.set_input_file_count(delta_manifest_entries)
    compaction_audit.set_total_cluster_memory_bytes(worker_task_mem)
    compaction_audit.set_hash_bucket_count(hash_bucket_count)

    logger.info(f"Hash bucket chunk size: {hash_bucket_chunk_size}")
    logger.info(f"Hash bucket count: {hash_bucket_count}")
    logger.info(f"Input uniform delta count: {len(rebatched_da_list)}")

    return rebatched_da_list, hash_bucket_count, high_watermark, require_multiple_rounds


def fit_input_deltas(
    input_deltas: List[Delta],
    cluster_resources: Dict[str, float],
    compaction_audit: CompactionSessionAuditInfo,
    hash_bucket_count: Optional[int],
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs,
) -> Tuple[List[DeltaAnnotated], int, HighWatermark, bool]:
    """
    This method tries to fit all the input deltas to run into the existing cluster. Contrary to
    'limit_input_deltas', it will not fail if the current cluster cannot run the compaction reliably.
    It is the responsibility of the caller to ensure that they pass enough resources for the job to execute.

    Note: There is a possibility that individual file could be very large, which makes the deltas non uniform.
    In such scenarios, it is advisable to allocate multiple vCPUs to the tasks to ensure parallelism.

    Args:
        input_deltas: The input deltas to be normalized.
        cluster_resources: Total available resources in the cluster.
        hash_bucket_count: The hash bucket count.
        deltacat_storage: An implementation of the DeltaCAT storage interface.

    Returns:
        Tuple of list of annotated deltas, recommended hash bucket count, high watermark,
            and whether multiple rounds are required (which is always False)
    """
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    worker_cpus = int(cluster_resources["CPU"])
    total_memory = float(cluster_resources["memory"])
    high_watermark = HighWatermark()
    annotated_input_da_list = []
    delta_bytes = 0
    total_files = 0

    if not input_deltas:
        raise AssertionError("No input deltas found!")

    for delta in input_deltas:
        manifest_entries = delta.manifest.entries
        position = delta.stream_position

        for entry in manifest_entries:
            delta_bytes += entry.meta.content_length

        total_files += len(manifest_entries)

        high_watermark.set(
            delta.locator.partition_locator,
            max(position, high_watermark.get(delta.locator.partition_locator)),
        )
        delta_annotated = DeltaAnnotated.of(delta)
        annotated_input_da_list.append(delta_annotated)

    # We assume that the cluster is capable of distributing all tasks
    # correctly. Hence, the correct in-memory size will be in the ratio of
    # in-disk size.
    def estimate_size(manifest_entry: ManifestEntry):
        return (manifest_entry.meta.content_length * 1.0 / delta_bytes) * total_memory

    # Assuming each CPU consumes equal amount of memory
    min_delta_bytes = total_memory / worker_cpus
    rebatched_da_list = DeltaAnnotated.rebatch(
        annotated_deltas=annotated_input_da_list,
        min_delta_bytes=min_delta_bytes,
        estimation_function=estimate_size,
    )

    # Recommended hash buckets based on the experiments performed
    # using S3 input for optimal throughput.
    if hash_bucket_count is None:
        hash_bucket_count = int(
            math.ceil(total_memory / MEMORY_TO_HASH_BUCKET_COUNT_RATIO)
        )

    compaction_audit.set_input_file_count(total_files)
    compaction_audit.set_input_size_bytes(delta_bytes)
    compaction_audit.set_total_cluster_memory_bytes(total_memory)
    compaction_audit.set_hash_bucket_count(hash_bucket_count)

    logger.info(
        f"Input delta bytes: {delta_bytes}, Total files: {total_files}, The worker_cpus: {worker_cpus}, "
        f" total_memory: {total_memory}, and hash_bucket_count: {hash_bucket_count}"
    )
    return rebatched_da_list, hash_bucket_count, high_watermark, False


def _discover_deltas(
    source_partition_locator: PartitionLocator,
    start_position_exclusive: Optional[int],
    end_position_inclusive: int,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
    list_deltas_kwargs: Optional[Dict[str, Any]] = {},
) -> List[Delta]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    kwargs = {**deltacat_storage_kwargs, **list_deltas_kwargs}

    deltas_list_result = deltacat_storage.list_partition_deltas(
        partition_like=source_partition_locator,
        first_stream_position=start_position_exclusive,
        last_stream_position=end_position_inclusive,
        ascending_order=True,
        include_manifest=True,
        **kwargs,
    )
    deltas = deltas_list_result.all_items()
    if not deltas:
        logger.warn(
            f"Couldn't find any deltas to "
            f"compact in delta stream position range "
            f"('{start_position_exclusive}', "
            f"'{end_position_inclusive}']. Source partition: "
            f"{source_partition_locator}"
        )
        return []

    if start_position_exclusive == deltas[0].stream_position:
        first_delta = deltas.pop(0)
        logger.info(
            f"Removed exclusive start delta w/ expected stream "
            f"position '{start_position_exclusive}' from deltas to "
            f"compact: {first_delta}"
        )
    logger.info(
        f"Count of deltas to compact in delta stream "
        f"position range ('{start_position_exclusive}', "
        f"'{end_position_inclusive}']: {len(deltas)}. Source "
        f"partition: '{source_partition_locator}'"
    )
    return deltas



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import functools
import logging
import os
import pathlib
from typing import Any, Dict, List, Optional, Set

import ray
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.metastats.model.stats_cluster_size_estimator import (
    StatsClusterSizeEstimator,
)
from deltacat.compute.metastats.stats import start_stats_collection
from deltacat.compute.metastats.utils.constants import (
    DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE,
    DEFAULT_JOB_RUN_TRACE_ID,
    HEAD_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO,
    MANIFEST_FILE_COUNT_PER_CPU,
    STATS_CLUSTER_R5_INSTANCE_TYPE,
    WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO,
)
from deltacat.compute.metastats.utils.io import read_cached_partition_stats
from deltacat.compute.metastats.utils.pyarrow_memory_estimation_function import (
    estimation_function,
)
from deltacat.compute.metastats.utils.ray_utils import replace_cluster_cfg_vars
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.compute.stats.models.delta_stats_cache_result import DeltaStatsCacheResult
from deltacat.compute.stats.utils.intervals import DeltaRange, merge_intervals
from deltacat.compute.stats.utils.io import get_deltas_from_range
from deltacat.constants import (
    BYTES_PER_GIBIBYTE,
    PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS,
)
from deltacat.storage import Delta, DeltaLocator, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.utils.performance import timed_invocation

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def collect_metastats(
    source_partition_locators: List[PartitionLocator],
    columns: Optional[List[str]] = None,
    file_count_per_cpu: Optional[int] = MANIFEST_FILE_COUNT_PER_CPU,
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    *args,
    **kwargs,
) -> Dict[str, Dict[int, DeltaStats]]:

    # TODO: Add CompactionEventDispatcher for metastats collection started event
    stats_res_all_partitions: Dict[str, Dict[int, DeltaStats]] = {}
    stats_res_obj_ref_all_partitions: Dict[str, ObjectRef] = {}
    for partition_locator in source_partition_locators:
        partition_id = partition_locator.partition_id
        if partition_locator.partition_values:
            partition_value_string = "_".join(partition_locator.partition_values)
        else:
            partition_value_string = f"no_partition_value_{partition_id}"
        partition_canonical_string = partition_locator.canonical_string()
        stats_res_obj_ref = collect_from_partition.remote(
            source_partition_locator=partition_locator,
            partition_value_string=partition_value_string,
            partition_canonical_string=partition_canonical_string,
            columns=columns,
            stat_results_s3_bucket=stat_results_s3_bucket,
            metastats_results_s3_bucket=metastats_results_s3_bucket,
            file_count_per_cpu=file_count_per_cpu,
            deltacat_storage=deltacat_storage,
            *args,
            **kwargs,
        )
        stats_res_obj_ref_all_partitions[partition_value_string] = stats_res_obj_ref
    for pv, stats_res_obj_ref in stats_res_obj_ref_all_partitions.items():
        stats_res_all_partitions[pv] = ray.get(stats_res_obj_ref)
    # TODO: Add CompactionEventDispatcher for metastats collection completed event

    logger.info(f"stats_res_all_partitions: {stats_res_all_partitions}")

    # For compaction result validation purpose only
    aggregate_partition_stats_for_validation: Dict[str, list] = {}
    for partition_val, delta_stream_range_set in stats_res_all_partitions.items():
        partition_stats_sum_row_count = 0
        partition_pyarrow_sum = 0
        for stream_pos, stats_column_result in delta_stream_range_set.items():
            for cs in stats_column_result.column_stats[0].manifest_stats.stats:
                partition_stats_sum_row_count += cs.get("rowCount")

            for stats in stats_column_result.get("column_stats"):
                partition_pyarrow_sum += stats.get("stats").get("pyarrowTableBytes")
            aggregate_partition_stats_for_validation[partition_val] = [
                partition_stats_sum_row_count,
                partition_pyarrow_sum,
            ]
        logger.info(
            f"partitions_stats_result for partition value: {partition_val}: rowCount: {partition_stats_sum_row_count}; pyarrowTableBytes: {partition_pyarrow_sum}"
        )
    return aggregate_partition_stats_for_validation

    # return stats_res_all_partitions


@ray.remote(num_cpus=1)
def collect_from_partition(
    source_partition_locator: PartitionLocator,
    partition_value_string,
    partition_canonical_string,
    delta_stream_position_range_set: Optional[Set[DeltaRange]] = None,
    columns: Optional[List[str]] = None,
    file_count_per_cpu: Optional[int] = MANIFEST_FILE_COUNT_PER_CPU,
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = None,
    *args,
    **kwargs,
) -> ObjectRef[Dict[int, DeltaStats]]:
    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}
    if not columns:
        columns = deltacat_storage.get_table_version_column_names(
            source_partition_locator.namespace,
            source_partition_locator.table_name,
            source_partition_locator.table_version,
        )
    deltas = _find_deltas(
        source_partition_locator, delta_stream_position_range_set, deltacat_storage
    )

    logger.info(f"Find {len(deltas)} deltas!")
    trace_id = DEFAULT_JOB_RUN_TRACE_ID
    if "trace_id" in kwargs:
        trace_id = kwargs.get("trace_id")
    else:
        logger.warning(
            f"No job run trace id specified, default to {DEFAULT_JOB_RUN_TRACE_ID}"
        )

    cpus_per_instance = DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE
    if cpus_per_instance in kwargs:
        cpus_per_instance = kwargs.get("cpus_per_instance")
    else:
        logger.info(
            f"Stats cluster CPUS per instance not specified, default to {DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE}"
        )

    stats_res_obj_ref = _start_all_stats_collection_from_deltas(
        deltas,
        partition_value_string,
        partition_canonical_string,
        columns,
        trace_id,
        file_count_per_cpu,
        cpus_per_instance,
        stat_results_s3_bucket,
        metastats_results_s3_bucket,
        deltacat_storage,
    )
    return stats_res_obj_ref


def _find_deltas(
    source_partition_locator: PartitionLocator,
    delta_stream_position_range_set: Optional[Set[DeltaRange]] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> List[Delta]:

    if delta_stream_position_range_set is None:
        delta_stream_position_range_set = {(None, None)}
    delta_range_lookup_pending: List[ObjectRef[List[Delta]]] = []

    for range_pair in merge_intervals(delta_stream_position_range_set):
        begin, end = range_pair
        promise: ObjectRef[List[Delta]] = get_deltas_from_range.remote(
            source_partition_locator, begin, end, deltacat_storage
        )
        delta_range_lookup_pending.append(promise)

    delta_list_by_ranges: List[List[Delta]] = ray.get(delta_range_lookup_pending)
    deltas = [delta for delta_list in delta_list_by_ranges for delta in delta_list]
    return deltas


def _start_all_stats_collection_from_deltas(
    deltas: List[Delta],
    partition_value_string: Optional[str],
    partition_canonical_string: Optional[str],
    columns: Optional[List[str]] = None,
    trace_id: Optional[str] = None,
    file_count_per_cpu: Optional[int] = MANIFEST_FILE_COUNT_PER_CPU,
    cpus_per_instance: Optional[int] = DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE,
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> Dict[int, DeltaStats]:

    delta_stats_compute_list: List[DeltaLocator] = []
    meta_stats_list_ready: List[DeltaLocator] = []
    meta_stats_list_to_compute: List[DeltaLocator] = []

    if stat_results_s3_bucket:
        found_columns_stats_map: Dict[
            int, List[DeltaStatsCacheResult]
        ] = read_cached_partition_stats(
            partition_canonical_string, stat_results_s3_bucket
        )

    delta_cache_res: List[DeltaStats] = []
    for delta in deltas:
        if found_columns_stats_map and delta.stream_position in found_columns_stats_map:
            cached_result = found_columns_stats_map[delta.stream_position]
            if cached_result.hits:
                delta_cache_res.append(cached_result.hits)
                meta_stats_list_ready.append(
                    cached_result.hits.column_stats[0].manifest_stats.delta_locator
                )

            if cached_result.misses:
                delta_locator: DeltaLocator = cached_result.misses.delta_locator
                delta_stats_compute_list.append(delta_locator)
                meta_stats_list_to_compute.append(delta_locator)
        else:
            delta_stats_compute_list.append(delta.locator)
            meta_stats_list_to_compute.append(delta.locator)

    logger.info(f"Collecting stats on {len(delta_stats_compute_list)} deltas!")
    delta_stats_compute_res: Dict[int, DeltaStats] = {}
    if delta_stats_compute_list:
        delta_stats_compute_res = _start_metadata_stats_collection(
            delta_stats_compute_list=delta_stats_compute_list,
            meta_stats_list_ready=meta_stats_list_ready,
            meta_stats_list_to_compute=meta_stats_list_to_compute,
            partition_value_string=partition_value_string,
            partition_canonical_string=partition_canonical_string,
            columns=columns,
            trace_id=trace_id,
            file_count_per_cpu=file_count_per_cpu,
            cpus_per_instance=cpus_per_instance,
            stat_results_s3_bucket=stat_results_s3_bucket,
            metastats_results_s3_bucket=metastats_results_s3_bucket,
            deltacat_storage=deltacat_storage,
        )

    delta_stream_range_stats: Dict[int, DeltaStats] = {}
    for delta_column_stats in delta_cache_res:
        assert (
            len(delta_column_stats.column_stats) > 0
        ), f"Expected columns of `{delta_column_stats}` to be non-empty"
        stream_position = delta_column_stats.column_stats[
            0
        ].manifest_stats.delta_locator.stream_position
        delta_stream_range_stats[stream_position] = delta_column_stats

    # stats collection result: if we have cached stats and missed column stats for same delta, stats collection for this delta is still needed
    # and the final result will use the newly collected stats for this delta.
    stats_collection_res: Dict[int, DeltaStats] = {
        **delta_stream_range_stats,
        **delta_stats_compute_res,
    }

    return stats_collection_res


def _start_metadata_stats_collection(
    delta_stats_compute_list: List[DeltaLocator],
    meta_stats_list_ready: List[DeltaLocator],
    meta_stats_list_to_compute: List[DeltaLocator],
    partition_value_string: Optional[str],
    partition_canonical_string: Optional[str],
    columns: Optional[List[str]] = None,
    trace_id: Optional[str] = None,
    file_count_per_cpu: Optional[int] = MANIFEST_FILE_COUNT_PER_CPU,
    cpus_per_instance: Optional[int] = DEFAULT_CPUS_PER_INSTANCE_R5_8XLARGE,
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> Dict[int, DeltaStats]:

    meta_stats_res_ready: Dict[int, int] = {}

    for delta_locator in meta_stats_list_ready:
        delta_meta_count = 0
        manifest = deltacat_storage.get_delta_manifest(delta_locator)
        delta = Delta.of(delta_locator, None, None, None, manifest)

        for entry in delta.manifest.entries:
            delta_meta_count += entry.meta.content_length
        meta_stats_res_ready[delta.stream_position] = delta_meta_count

    first_delta_locator = (
        meta_stats_list_ready[0]
        if meta_stats_list_ready
        else meta_stats_list_to_compute[0]
    )
    manifest = deltacat_storage.get_delta_manifest(first_delta_locator)
    content_type = manifest.meta.content_type
    content_encoding = manifest.meta.content_type

    meta_stats_to_compute: Dict[int, int] = {}
    manifest_file_count_to_compute: Dict[int, int] = {}

    for delta_locator in meta_stats_list_to_compute:
        delta_meta_count = 0
        manifest = deltacat_storage.get_delta_manifest(delta_locator)
        delta = Delta.of(delta_locator, None, None, None, manifest)
        file_count = len(delta.manifest.entries)
        manifest_file_count_to_compute[delta.stream_position] = file_count
        for entry in delta.manifest.entries:
            delta_meta_count += entry.meta.content_length
        meta_stats_to_compute[delta.stream_position] = delta_meta_count

    batched_delta_stats_compute_list = _batch_deltas(
        delta_stats_compute_list,
        file_count_per_cpu,
        cpus_per_instance,
        deltacat_storage,
        content_type,
        content_encoding,
    )

    # out_cluster_cfg = _setup_stats_cluster(min_workers,
    #                                        partition_value_string,
    #                                        trace_id,
    #                                        cpus_per_instance)
    out_cluster_cfg = None
    delta_stats_res: Dict[int, DeltaStats] = _start_stats_cluster(
        out_cluster_cfg,
        batched_delta_stats_compute_list,
        columns,
        stat_results_s3_bucket,
        metastats_results_s3_bucket,
        deltacat_storage,
        partition_canonical_string,
    )

    return delta_stats_res


def _start_stats_cluster(
    out_cluster_cfg: str,
    batched_delta_stats_compute_list: List[DeltaAnnotated],
    columns: List[str],
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
    partition_val: Optional[str] = "partition_val",
):
    # ray_up_latency = timed_invocation(
    #     func=ray_up,
    #     cluster_cfg=out_cluster_cfg
    # )
    # logger.info(f"ray_up_latency: {partition_val}:{ray_up_latency}")

    # head_node_ip = get_head_node_ip(out_cluster_cfg)
    # client = ray_init(head_node_ip, 10001)
    # with client:
    delta_stream_range_stats, stats_collection_latency = timed_invocation(
        func=start_stats_collection,
        batched_delta_stats_compute_list=batched_delta_stats_compute_list,
        columns=columns,
        stat_results_s3_bucket=stat_results_s3_bucket,
        metastats_results_s3_bucket=metastats_results_s3_bucket,
        deltacat_storage=deltacat_storage,
    )
    logger.info(
        f"actual_stats_collection_latency: {partition_val}: {stats_collection_latency}"
    )
    # client.disconnect()
    # ray_down(out_cluster_cfg)
    # clean_up_cluster_cfg_file(out_cluster_cfg)
    return delta_stream_range_stats


def _estimate_cpus_needed(
    meta_stats_to_compute,
    memory_gb_per_cpu,
    file_count_per_cpu,
    manifest_file_count_to_compute,
    partition_val,
):
    content_length_sum = 0
    for val in meta_stats_to_compute.values():
        content_length_sum += val
    manifest_file_count_sum = 0
    for val in manifest_file_count_to_compute.values():
        manifest_file_count_sum += val
    estimated_memory_bytes_needed = (
        content_length_sum * PYARROW_INFLATION_MULTIPLIER_ALL_COLUMNS
    )
    estimated_memory_gib_needed = estimated_memory_bytes_needed / BYTES_PER_GIBIBYTE

    logger.info(
        f"estimated_memory_gib_needed: {partition_val} : {estimated_memory_gib_needed}"
    )
    logger.info(f"manifest_file_count_sum: {partition_val} : {manifest_file_count_sum}")

    memory_per_cpu_available = memory_gb_per_cpu * (
        1 - WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO
    )
    estimator = StatsClusterSizeEstimator.of(
        memory_per_cpu_available,
        file_count_per_cpu,
        estimated_memory_gib_needed,
        manifest_file_count_sum,
    )
    min_cpus = StatsClusterSizeEstimator.estimate_cpus_needed(estimator)
    return min_cpus


def _batch_deltas(
    delta_stats_compute_list,
    file_count_per_cpu,
    cpu_per_instance,
    deltacat_storage,
    content_type,
    content_encoding,
) -> List[DeltaAnnotated]:
    worker_node_mem = (
        cpu_per_instance
        * (1 - WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO)
        * BYTES_PER_GIBIBYTE
    )
    delta_list = []

    estimate_based_on_content_length = functools.partial(
        estimation_function,
        content_type=content_type,
        content_encoding=content_encoding,
    )

    for delta_locator in delta_stats_compute_list:
        manifest = deltacat_storage.get_delta_manifest(delta_locator)
        delta = Delta.of(delta_locator, None, None, None, manifest)
        delta_annotated = DeltaAnnotated.of(delta)
        delta_list.append(delta_annotated)

    rebatched_da_list = DeltaAnnotated.rebatch(
        delta_list,
        worker_node_mem,
        file_count_per_cpu,
        estimate_based_on_content_length,
    )

    logger.info(f"Rebatched_delta_list_length: {len(rebatched_da_list)}")

    return rebatched_da_list


def _setup_stats_cluster(
    min_workers, partition_value_string, trace_id, cpus_per_instance
):
    stats_cluster_instance_type = (
        int(cpus_per_instance // 4)
        if cpus_per_instance
        else STATS_CLUSTER_R5_INSTANCE_TYPE
    )
    stats_cluster_instance_type_str = f"r5.{stats_cluster_instance_type}xlarge".strip()
    parent_dir_path = pathlib.Path(__file__).parent.resolve()
    in_cfg = os.path.join(parent_dir_path, "config", "stats_cluster_example.yaml")
    out_cluster_cfg_file_path = replace_cluster_cfg_vars(
        partition_canonical_string=partition_value_string,
        trace_id=trace_id,
        file_path=in_cfg,
        min_workers=min_workers,
        head_type=stats_cluster_instance_type_str,
        worker_type=stats_cluster_instance_type_str,
        head_object_store_memory_pct=HEAD_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO * 100,
        worker_object_store_memory_pct=WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO
        * 100,
    )

    return out_cluster_cfg_file_path



from typing import NamedTuple, List
from deltacat.compute.compactor.model.materialize_result import MaterializeResult

import numpy as np


class MergeResult(NamedTuple):
    materialize_results: List[MaterializeResult]
    deduped_record_count: np.int64
    peak_memory_usage_bytes: np.double
    telemetry_time_in_seconds: np.double
    task_completed_at: np.double



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Optional

from deltacat.compute.stats.models.delta_stats import DeltaStats, DeltaStatsCacheMiss


class DeltaStatsCacheResult(dict):
    """A helper class containing the results from a cache query.

    Stats are fetched and cached at the column level, and each column may represent one
    or more manifest entries.
    """

    @staticmethod
    def of(
        hits: Optional[DeltaStats], misses: Optional[DeltaStatsCacheMiss]
    ) -> DeltaStatsCacheResult:
        cds = DeltaStatsCacheResult()
        cds["hits"] = hits
        cds["misses"] = misses
        return cds

    @property
    def hits(self) -> Optional[DeltaStats]:
        """Retrieve stats that were found in the cache

        `hits` represents a DeltaStats object that contains dataset-wide statistics across
        many of its tables (or manifest entries) and is composed of one or more column-wide
        DeltaColumnStats.

        Returns:
            A delta wide stats container
        """
        return self["hits"]

    @property
    def misses(self) -> Optional[DeltaStatsCacheMiss]:
        """Retrieve stats that were missing from the cache

        `misses` represents a DeltaStatsCacheMiss object that contains a list of
        column names that were not found in the file system (ex: S3) and a `delta_locator`
        as a reference to the delta metadata tied to the missing dataset columns.

        Returns:
            A tuple with metadata regarding the cache miss
        """
        return self["misses"]



import logging
import time
from typing import Any, Callable, Dict, List

import ray

from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def node_resource_keys(
    filter_fn: Callable[[Dict[str, Any]], bool] = lambda n: True
) -> List[str]:
    """Get all Ray resource keys for cluster nodes that pass the given filter
    as a list of strings of the form: "node:{node_resource_name}". The returned
    keys can be used to place tasks or actors on that node via:
    `foo.options(resources={node_resource_keys()[0]: 0.01}).remote()`

    Returns all cluster node resource keys by default."""
    keys = []
    node_dict = ray.nodes()
    if node_dict:
        for node in ray.nodes():
            if filter_fn(node):
                for key in node["Resources"].keys():
                    if key.startswith("node:"):
                        keys.append(key)
    else:
        raise ValueError("No node dictionary found on current node.")
    return keys


def current_node_resource_key() -> str:
    """Get the Ray resource key for the current node as a string of the form:
    "node:{node_resource_name}". The returned key can be used to place tasks or
    actors on that node via:
    `foo.options(resources={get_current_node_resource_key(): 0.01}).remote()`
    """
    current_node_id = ray.get_runtime_context().node_id.hex()
    keys = node_resource_keys(lambda n: n["NodeID"] == current_node_id)
    assert (
        len(keys) <= 1
    ), f"Expected <= 1 keys for the current node, but found {len(keys)}"
    return keys[0] if len(keys) == 1 else None


def is_node_alive(node: Dict[str, Any]) -> bool:
    """Takes a node from `ray.nodes()` as input. Returns True if the node is
    alive, and False otherwise."""
    return node["Alive"]


def live_node_count() -> int:
    """Returns the number of nodes in the cluster that are alive."""
    return sum(1 for n in ray.nodes() if is_node_alive(n))


def live_node_waiter(min_live_nodes: int, poll_interval_seconds: float = 0.5) -> None:
    """Waits until the given minimum number of live nodes are present in the
    cluster. Checks the current number of live nodes every
    `poll_interval_seconds`."""
    live_nodes = live_node_count()
    while live_nodes < min_live_nodes:
        live_nodes = live_node_count()
        logger.info(f"Waiting for Live Nodes: {live_nodes}/{min_live_nodes}")
        time.sleep(poll_interval_seconds)


def live_node_resource_keys() -> List[str]:
    """Get Ray resource keys for all live cluster nodes as a list of strings of
    the form: "node:{node_resource_name}". The returned keys can be used to
    place tasks or actors on that node via:
    `foo.options(resources={node_resource_keys()[0]: 0.01}).remote()`"""
    return node_resource_keys(lambda n: is_node_alive(n))


def other_live_node_resource_keys() -> List[str]:
    """Get Ray resource keys for all live cluster nodes except the current node
    as a list of strings of the form: "node:{node_resource_name}". The returned
    keys can be used to place tasks or actors on that node via:
    `foo.options(resources={node_resource_keys()[0]: 0.01}).remote()`

    For example, invoking this function from your Ray application driver on the
    head node returns the resource keys of all live worker nodes."""
    current_node_id = ray.get_runtime_context().node_id.hex()
    return node_resource_keys(
        lambda n: n["NodeID"] != current_node_id and is_node_alive(n)
    )


def other_node_resource_keys() -> List[str]:
    """Get Ray resource keys for all cluster nodes except the current node
    as a list of strings of the form: "node:{node_resource_name}". The returned
    keys can be used to place tasks or actors on that node via:
    `foo.options(resources={node_resource_keys()[0]: 0.01}).remote()`

    For example, invoking this function from your Ray application driver on the
    head node returns the resource keys of all worker nodes."""
    current_node_id = ray.get_runtime_context().node_id.hex()
    return node_resource_keys(lambda n: n["NodeID"] != current_node_id)


def cluster_cpus() -> int:
    """Returns the current cluster's total number of CPUs as an integer."""
    cpus = ray.cluster_resources().get("CPU")
    return int(cpus) if cpus is not None else 0


def available_cpus() -> int:
    """Returns the current cluster's number of available CPUs as an integer."""
    cpus = ray.available_resources()["CPU"]
    return int(cpus) if cpus is not None else 0


def log_cluster_resources() -> None:
    """Logs the cluster's total and available resources."""
    logger.info(f"Available Resources: {ray.available_resources()}")
    logger.info(f"Cluster Resources: {ray.cluster_resources()}")
    logger.info(f"Cluster Nodes: {ray.nodes()}")


def get_current_ray_task_id() -> str:
    return ray.get_runtime_context().get_task_id()


def get_current_ray_worker_id() -> str:
    return ray.get_runtime_context().worker.core_worker.get_worker_id()



import unittest
import sqlite3
import ray
import os
from unittest.mock import patch
import deltacat.tests.local_deltacat_storage as ds
from deltacat.types.media import ContentType
from deltacat.compute.compactor_v2.compaction_session import compact_partition
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.utils.common import current_time_ms
from deltacat.tests.test_utils.pyarrow import stage_partition_from_file_paths


class TestCompactionSession(unittest.TestCase):
    """
    This class adds specific tests that aren't part of the parametrized test suite.
    """

    DB_FILE_PATH = f"{current_time_ms()}.db"
    NAMESPACE = "compact_partition_v2_namespace"

    @classmethod
    def setUpClass(cls):
        ray.init(local_mode=True, ignore_reinit_error=True)

        con = sqlite3.connect(cls.DB_FILE_PATH)
        cur = con.cursor()
        cls.kwargs = {ds.SQLITE_CON_ARG: con, ds.SQLITE_CUR_ARG: cur}
        cls.deltacat_storage_kwargs = {ds.DB_FILE_PATH_ARG: cls.DB_FILE_PATH}

        super().setUpClass()

    @classmethod
    def doClassCleanups(cls) -> None:
        os.remove(cls.DB_FILE_PATH)

    @patch("deltacat.compute.compactor_v2.compaction_session.rcf")
    @patch("deltacat.compute.compactor_v2.compaction_session.s3_utils")
    def test_compact_partition_when_no_input_deltas_to_compact(self, s3_utils, rcf_url):
        # setup
        rcf_url.read_round_completion_file.return_value = None
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["test"], **self.deltacat_storage_kwargs
        )
        source_partition = ds.commit_partition(
            staged_source, **self.deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **self.deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **self.deltacat_storage_kwargs
        )

        # action
        rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": "test_bucket",
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": self.deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 1,
                    "last_stream_position_to_compact": source_partition.stream_position,
                    "list_deltas_kwargs": {
                        **self.deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": [],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_partition.locator,
                }
            )
        )

        # verify that no RCF is written
        self.assertIsNone(rcf_url)
